@@ -96,6 +96,9 @@ Rcpp::NumericVector inv_geotransform(const std::vector<double> gt);
 Rcpp::IntegerMatrix get_pixel_line(const Rcpp::NumericMatrix xy,
 		const std::vector<double> gt);
 
+bool buildVRT(std::string vrt_filename, Rcpp::CharacterVector input_rasters,
+		Rcpp::Nullable<Rcpp::CharacterVector> cl_arg);
+		
 Rcpp::DataFrame _combine(Rcpp::CharacterVector src_files,
 		Rcpp::CharacterVector var_names,
 		std::vector<int> bands, 
@@ -114,16 +117,26 @@ bool _dem_proc(std::string mode,
 
 bool fillNodata(std::string filename, int band, std::string mask_file,
 		double max_dist, int smooth_iterations);
+
+bool _polygonize(std::string src_filename, int src_band,
+		std::string out_dsn, std::string out_layer, std::string fld_name,
+		std::string mask_file, bool nomask, int connectedness);
+		
+bool _rasterize(std::string src_dsn, std::string dst_filename,
+		Rcpp::CharacterVector cl_arg);
 		
 bool sieveFilter(std::string src_filename, int src_band,
 		std::string dst_filename, int dst_band,
 		int size_threshold, int connectedness,
 		std::string mask_filename , int mask_band,
 		Rcpp::Nullable<Rcpp::CharacterVector> options);
+
+bool translate(std::string src_filename, std::string dst_filename,
+		Rcpp::Nullable<Rcpp::CharacterVector> cl_arg);
 		
-bool warp(Rcpp::CharacterVector src_files, std::string dst_filename,
-		Rcpp::CharacterVector t_srs, 
-		Rcpp::Nullable<Rcpp::CharacterVector> arg_list);
+bool warp(std::vector<std::string> src_files, std::string dst_filename,
+		std::string t_srs, 
+		Rcpp::Nullable<Rcpp::CharacterVector> cl_arg);
 		
 Rcpp::IntegerMatrix createColorRamp(int start_index,
 		Rcpp::IntegerVector start_color,
